@@ -14,6 +14,7 @@ See the License for the specific language governing permissions and
 limitations under the License.
 """
 
+import os
 import pathlib
 import re
 import warnings
@@ -32,8 +33,11 @@ def _get_workspace_dir_name() -> pathlib.Path:
         arch = "_".join(sorted(set(re.findall(r"compute_(\d+)", "".join(flags)))))
     except Exception:
         arch = "noarch"
+    flashinfer_base = os.getenv(
+        "FLASHINFER_WORKSPACE_BASE", pathlib.Path.home().as_posix()
+    )
     # e.g.: $HOME/.cache/flashinfer/75_80_89_90/
-    return pathlib.Path.home() / ".cache" / "flashinfer" / arch
+    return pathlib.Path(flashinfer_base) / ".cache" / "flashinfer" / arch
 
 
 # use pathlib
