@@ -333,14 +333,17 @@ def get_sampling_module():
 
         @register_custom_op(
             "flashinfer::chain_speculative_sampling",
-            mutates_args=("output_accepted_token_num", "output_emitted_token_num"),
+            mutates_args=(
+                "output_accepted_token_num",
+                "output_emitted_draft_token_num",
+            ),
         )
         def chain_speculative_sampling(
             draft_probs: torch.Tensor,
             draft_token_ids: torch.Tensor,
             target_probs: torch.Tensor,
             output_accepted_token_num: torch.Tensor,
-            output_emitted_token_num: torch.Tensor,
+            output_emitted_draft_token_num: torch.Tensor,
             deterministic: bool,
             generator: Optional[torch.Generator],
         ) -> torch.Tensor:
@@ -349,7 +352,7 @@ def get_sampling_module():
             draft_token_ids = draft_token_ids.int()
             target_probs = target_probs.float()
             output_accepted_token_num = output_accepted_token_num.int()
-            output_emitted_token_num = output_emitted_token_num.int()
+            output_emitted_draft_token_num = output_emitted_draft_token_num.int()
             b, n = draft_token_ids.shape
             output_token_ids = torch.empty((b, n + 1), dtype=torch.int32, device=device)
             module.chain_speculative_sampling.default(
@@ -358,7 +361,7 @@ def get_sampling_module():
                 target_probs,
                 output_token_ids,
                 output_accepted_token_num,
-                output_emitted_token_num,
+                output_emitted_draft_token_num,
                 deterministic,
                 generator,
             )
@@ -370,7 +373,7 @@ def get_sampling_module():
             draft_token_ids: torch.Tensor,
             target_probs: torch.Tensor,
             output_accepted_token_num: torch.Tensor,
-            output_emitted_token_num: torch.Tensor,
+            output_emitted_draft_token_num: torch.Tensor,
             deterministic: bool,
             generator: Optional[torch.Generator],
         ) -> torch.Tensor:
@@ -1130,7 +1133,7 @@ def chain_speculative_sampling(
     draft_token_ids,
     target_probs,
     maybe_output_accepted_token_num: Optional[torch.Tensor] = None,
-    maybe_output_emitted_token_num: Optional[torch.Tensor] = None,
+    maybe_output_emitted_draft_token_num: Optional[torch.Tensor] = None,
     deterministic: bool = True,
     generator: Optional[torch.Generator] = None,
 ) -> torch.Tensor:
@@ -1158,8 +1161,10 @@ def chain_speculative_sampling(
         It only evaluates the alignment of draft model and target model.
         Shape: ``(batch_size)``
         If specified, the number of accepted token number will be added to this tensor inplace. Default is ``None``.
-    maybe_output_emitted_token_num: Optional[torch.Tensor]
-        The number of tokens that are finally emitted/generated for each request.
+    maybe_output_emitted_draft_token_num: Optional[torch.Tensor]
+        The number of draft tokens that are finally emitted for each request. Does not include
+        the bonus token. (Thus the total number of tokens sampled for a given request is
+        output_emitted_draft_token_num + 1).
         Shape: ``(batch_size)``
         If specified, the number of emitted token number will be added to this tensor inplace. Default is ``None``.
     deterministic: bool
@@ -1182,8 +1187,10 @@ def chain_speculative_sampling(
         satisfy the probability requirement r < p/q.
         It only evaluates the alignment of draft model and target model.
         Shape: ``(batch_size)``
-    output_emitted_token_num: torch.Tensor
-        The number of tokens that are finally emitted/generated for each request.
+    output_emitted_draft_token_num: torch.Tensor
+        The number of draft tokens that are finally emitted for each request. Does not include
+        the bonus token. (Thus the total number of tokens sampled for a given request is
+        output_emitted_draft_token_num + 1).
         Shape: ``(batch_size)``
 
     Examples
@@ -1200,7 +1207,7 @@ def chain_speculative_sampling(
     >>> # token 1 was sampled from draft model for the second token
     >>> draft_token_ids = torch.tensor([[2, 1]], dtype=torch.int32).to(0)
     >>> target_probs = torch.tensor([[[0.0, 0.1, 0.6, 0.3], [1.0, 0.0, 0.0, 0.0], [0.7, 0.1, 0.1, 0.1]]]).to(0)
-    >>> output_token_ids, output_accepted_token_num, output_accepted_token_num =\
+    >>> output_token_ids, output_accepted_token_num, output_emitted_draft_token_num =\
     ...     flashinfer.sampling.chain_speculative_sampling(
     ...         draft_probs, draft_token_ids, target_probs)
     >>> # the first token is accepted, the second token is rejected and sampled from the difference
@@ -1209,7 +1216,7 @@ def chain_speculative_sampling(
     tensor([[ 2,  0, -1]], device='cuda:0', dtype=torch.int32)
     >>> output_accepted_token_num
     tensor([1], device='cuda:0')
-    >>> output_emitted_token_num
+    >>> output_emitted_draft_token_num
     tensor([1], device='cuda:0')
     """
     b = draft_probs.size(0)
@@ -1218,17 +1225,17 @@ def chain_speculative_sampling(
         output_accepted_token_num = torch.zeros(b, dtype=torch.int32, device=dev)
     else:
         output_accepted_token_num = maybe_output_accepted_token_num
-    if maybe_output_emitted_token_num is None:
-        output_emitted_token_num = torch.zeros(b, dtype=torch.int32, device=dev)
+    if maybe_output_emitted_draft_token_num is None:
+        output_emitted_draft_token_num = torch.zeros(b, dtype=torch.int32, device=dev)
     else:
-        output_emitted_token_num = maybe_output_emitted_token_num
+        output_emitted_draft_token_num = maybe_output_emitted_draft_token_num
     output_token_ids = get_sampling_module().chain_speculative_sampling(
         draft_probs,
         draft_token_ids,
         target_probs,
         output_accepted_token_num,
-        output_emitted_token_num,
+        output_emitted_draft_token_num,
         deterministic,
         generator,
     )
-    return output_token_ids, output_accepted_token_num, output_emitted_token_num
+    return output_token_ids, output_accepted_token_num, output_emitted_draft_token_num
