@@ -346,7 +346,7 @@ class DeepseekV2AttentionMatAbsorbDecode(nn.Module):
                 dev_id
             )
             wrapper = flashinfer.BatchDecodeMlaWithPagedKVCacheWrapper(
-                workspace_buffer, 
+                workspace_buffer,
                 use_cuda_graph=True,
                 use_tensor_cores=True,
                 paged_kv_indptr_buffer=kv_indptr,
@@ -365,12 +365,14 @@ class DeepseekV2AttentionMatAbsorbDecode(nn.Module):
                 data_type=q_kv_dtype,
                 q_data_type=q_kv_dtype,
             )
-            
+
             g = torch.cuda.CUDAGraph()
             with torch.cuda.graph(g):
-                attn_output = wrapper.run(q_nope, q_pe, paged_ckv_cache, paged_kpe_cache)
+                attn_output = wrapper.run(
+                    q_nope, q_pe, paged_ckv_cache, paged_kpe_cache
+                )
             g.replay()
-        
+
         # output ~ [bsz, 5120]
         output = torch.matmul(
             attn_output.to(self.W_UV_O.dtype).reshape(
